from datetime import datetime
from enum import Enum
import ftplib
import io
import itertools
from os import PathLike
from pathlib import Path
from typing import Any, Iterable, List, TextIO, Union

import geopandas
from geopandas import GeoDataFrame
import pandas
from pandas import DataFrame, Series
import typepigeon

from stormevents.nhc.storms import nhc_storms

ATCF_RECORD_START_YEAR = 1850

# suppress `SettingWithCopyWarning`
pandas.options.mode.chained_assignment = None

# https://www.nrlmry.navy.mil/atcf_web/docs/database/new/abrdeck.html
ATCF_FIELDS = {
    # BASIN - basin, e.g. WP, IO, SH, CP, EP, AL, SL
    'BASIN': 'basin',
    # CY - annual cyclone number: 1 through 99
    'CY': 'storm_number',
    # YYYYMMDDHH - Warning Date-Time-Group: 0000010100 through 9999123123. (note, 4 digit year)
    'YYYYMMDDHH': 'datetime',
    # TECHNUM/MIN - objective technique sorting number, minutes for best track: 00 - 99
    'TECHNUM/MIN': 'record_type_number',
    # TECH - acronym for each objective technique or CARQ or WRNG, BEST for best track.
    'TECH': 'record_type',
    # TAU - forecast period: -24 through 240 hours, 0 for best-track, negative taus used for CARQ and WRNG records.
    'TAU': 'forecast_hours',
    # LatN/S - Latitude (tenths of degrees) for the DTG: 0 through 900, N/S is the hemispheric index.
    'LatN/S': 'latitude',
    # LonE/W - Longitude (tenths of degrees) for the DTG: 0 through 1800, E/W is the hemispheric index.
    'LonE/W': 'longitude',
    # VMAX - Maximum sustained wind speed in knots: 0 through 300.
    'VMAX': 'max_sustained_wind_speed',
    # MSLP - Minimum sea level pressure, 1 through 1100 MB.
    'MSLP': 'central_pressure',
    # TY - Level of tc development: DB - disturbance, TD - tropical depression, TS - tropical storm, TY - typhoon, ST - super typhoon, TC - tropical cyclone, HU - hurricane, SD - subtropical depression, SS - subtropical storm, EX - extratropical systems, IN - inland, DS - dissipating, LO - low, WV - tropical wave, ET - extrapolated, XX - unknown.
    'TY': 'development_level',
    # RAD - Wind intensity (kts) for the radii defined in this record: 34, 50, 64.
    'RAD': 'isotach_radius',
    # WINDCODE - Radius code: AAA - full circle, QQQ - quadrant (NNQ, NEQ, EEQ, SEQ, SSQ, SWQ, WWQ, NWQ)
    'WINDCODE': 'isotach_quadrant_code',
    # RAD1 - If full circle, radius of specified wind intensity, If semicircle or quadrant, radius of specified wind intensity of circle portion specified in radius code. 0 - 1200 nm.
    'RAD1': 'isotach_radius_for_NEQ',
    # RAD2 - If full circle this field not used, If semicicle, radius (nm) of specified wind intensity for semicircle not specified in radius code, If quadrant, radius (nm) of specified wind intensity for 2nd quadrant (counting clockwise from quadrant specified in radius code). 0 through 1200 nm.
    'RAD2': 'isotach_radius_for_SEQ',
    # RAD3 - If full circle or semicircle this field not used, If quadrant, radius (nm) of specified wind intensity for 3rd quadrant (counting clockwise from quadrant specified in radius code). 0 through 1200 nm.
    'RAD3': 'isotach_radius_for_NWQ',
    # RAD4 - If full circle or semicircle this field not used, If quadrant, radius (nm) of specified wind intensity for 4th quadrant (counting clockwise from quadrant specified in radius code). 0 through 1200 nm.
    'RAD4': 'isotach_radius_for_SWQ',
    # RADP - pressure in millibars of the last closed isobar, 900 - 1050 mb.
    'RADP': 'background_pressure',
    # RRP - radius of the last closed isobar in nm, 0 - 9999 nm.
    'RRP': 'radius_of_last_closed_isobar',
    # MRD - radius of max winds, 0 - 999 nm.
    'MRD': 'radius_of_maximum_winds',
    # GUSTS - gusts, 0 through 995 kts.
    'GUSTS': 'gust_speed',
    # EYE - eye diameter, 0 through 999 nm.
    'EYE': 'eye_diameter',
    # SUBREGION - subregion code: A - Arabian Sea, B - Bay of Bengal, C - Central Pacific, E - Eastern Pacific, L - Atlantic, P - South Pacific (135E - 120W), Q - South Atlantic, S - South IO (20E - 135E), W - Western Pacific
    'SUBREGION': 'subregion_code',
    # MAXSEAS - max seas: 0 through 999 ft.
    'MAXSEAS': 'maximum_wave_height',
    # INITIALS - Forecaster's initials, used for tau 0 WRNG, up to 3 chars.
    'INITIALS': 'forecaster_initials',
    # DIR - storm direction in compass coordinates, 0 - 359 degrees.
    'DIR': 'direction',
    # SPEED - storm speed, 0 - 999 kts.
    'SPEED': 'speed',
    # STORMNAME - literal storm name, NONAME or INVEST. TCcyx used pre-1999, where: cy = Annual cyclone number 01 through 99, x = Subregion code: W, A, B, S, P, C, E, L, Q.
    'STORMNAME': 'name',
    # user data section as indicated by USERDEFINED parameter.
    'USERDEFINED': 'extra_values',
}

FORT_22_FIELDS = {
    # Time Record number in column 29. There can be multiple lines for a given time record depending on the number of isotachs reported in the ATCF File
    'RECORD': 'record_number',
    # number of isotachs reported in the ATCF file for the corresponding Time record.
    'ISOTACHS': 'num_isotachs',
    # Columns 31-34 indicate the selection of radii for that particular isotach. 0 indicates do not use this radius, and 1 indicates use this radius and corresponding wind speed.
    'ISOTACHSEL1': 'isotach_select_1',
    'ISOTACHSEL2': 'isotach_select_2',
    'ISOTACHSEL3': 'isotach_select_3',
    'ISOTACHSEL4': 'isotach_select_4',
    # Columns 35-38 are the designated Rmax values computed for each of the quadrants selected for each particular isotach.
    'RMAXQUADRANT1': 'radius_of_maximum_winds_quadrant_1',
    'RMAXQUADRANT2': 'radius_of_maximum_winds_quadrant_2',
    'RMAXQUADRANT3': 'radius_of_maximum_winds_quadrant_3',
    'RMAXQUADRANT4': 'radius_of_maximum_winds_quadrant_4',
    # Column 39 is the Holland B parameter computed using the formulas outlines in the Holland paper, and implemented using the aswip program.
    'HOLLANDB': 'holland_b',
    # Column 40-43 is the quadrant-varying Holland B parameter
    'HOLLANDB1': 'holland_b_quadrant_1',
    'HOLLANDB2': 'holland_b_quadrant_2',
    'HOLLANDB3': 'holland_b_quadrant_3',
    'HOLLANDB4': 'holland_b_quadrant_4',
    # Column 44-47 are the quadrant-varying Vmax calculated at the top of the planetary boundary (a wind reduction factor is applied to reduce the wind speed at the boundary to the 10-m surface)
    'VMAX1': 'max_sustained_wind_speed_1',
    'VMAX2': 'max_sustained_wind_speed_2',
    'VMAX3': 'max_sustained_wind_speed_3',
    'VMAX4': 'max_sustained_wind_speed_4',
}

EXTRA_ATCF_FIELDS = {
    # DEPTH - system depth, D-deep, M-medium, S-shallow, X-unknown
    'DEPTH': 'depth_code',
    # SEAS - Wave height for radii defined in SEAS1-SEAS4, 0-99 ft.
    'SEAS': 'isowave',
    # SEASCODE - Radius code: AAA - full circle,  QQQ - quadrant (NNQ, NEQ, EEQ, SEQ, SSQ, SWQ, WWQ, NWQ)
    'SEASCODE': 'isowave_quadrant_code',
    # SEAS1 - first quadrant seas radius as defined by SEASCODE, 0 through 999 nm.
    'SEAS1': 'isowave_radius_for_NEQ',
    # SEAS2 - second quadrant seas radius as defined by SEASCODE, 0 through 999 nm.
    'SEAS2': 'isowave_radius_for_SEQ',
    # SEAS3 - third quadrant seas radius as defined by SEASCODE, 0 through 999 nm.
    'SEAS3': 'isowave_radius_for_NWQ',
    # SEAS4 - fourth quadrant seas radius as defined by SEASCODE, 0 through 999 nm.
    'SEAS4': 'isowave_radius_for_SWQ',
    # user data section as indicated by USERDEFINED parameter.
    'USERDEFINED': 'extra_values',
}


def atcf_files(
    file_deck: 'ATCF_FileDeck' = None, mode: 'ATCF_Mode' = None, year: int = None
) -> List[str]:
    if file_deck is None:
        return list(
            itertools.chain(
                *(
                    atcf_files(file_deck=file_deck.value, mode=mode, year=year)
                    for file_deck in ATCF_FileDeck
                )
            )
        )

    if mode is None:
        return list(
            itertools.chain(
                *(
                    atcf_files(file_deck=file_deck, mode=mode.value, year=year)
                    for mode in ATCF_Mode
                )
            )
        )

    if not isinstance(file_deck, ATCF_FileDeck):
        file_deck = typepigeon.convert_value(file_deck, ATCF_FileDeck)

    if not isinstance(mode, ATCF_Mode):
        mode = typepigeon.convert_value(mode, ATCF_Mode)

    if mode == ATCF_Mode.historical and year is None or isinstance(year, Iterable):
        if year is None:
            year = range(ATCF_RECORD_START_YEAR, datetime.today().year + 1)
        return list(
            itertools.chain(
                *(atcf_files(file_deck=file_deck, mode=mode, year=entry) for entry in year)
            )
        )

    url = atcf_url(file_deck=file_deck, mode=mode, year=year)
    hostname, directory = url.split('/', 3)[2:]
    ftp = ftplib.FTP(hostname.replace('ftp://', ''), 'anonymous', '')

    filenames = [
        filename
        for filename, metadata in ftp.mlsd(directory)
        if metadata['type'] == 'file' and filename[0] == file_deck.value
    ]

    filenames = sorted((filename for filename in filenames), reverse=True)

    if year is not None:
        filenames = [filename for filename in filenames if str(year) in filename]

    return [url + filename for filename in filenames]


class ATCF_FileDeck(Enum):
    """
    These formats specified by the Automated Tropical Cyclone Forecast (ATCF) System.
    The contents of each type of data file is described at http://hurricanes.ral.ucar.edu/realtime/
    """

    ADVISORY = 'a'
    BEST = 'b'  # "best track"
    FIXED = 'f'  # https://www.nrlmry.navy.mil/atcf_web/docs/database/new/newfdeck.txt


class ATCF_Mode(Enum):
    historical = 'ARCHIVE'
    realtime = 'aid_public'


class ATCF_RecordType(Enum):
    best = 'BEST'
    ofcl = 'OFCL'
    ofcp = 'OFCP'
    hmon = 'HMON'
    carq = 'CARQ'
    hwrf = 'HWRF'


def get_atcf_entry(
    year: int, basin: str = None, storm_number: int = None, storm_name: str = None,
) -> Series:
    storms = nhc_storms(year=year)

    if storm_name is None and (basin is None and storm_number is None):
        raise ValueError('need either storm name + year OR basin + storm number + year')

    if basin is not None:
        storms = storms[storms['basin'].str.contains(basin.upper())]
    if storm_number is not None:
        storms = storms[storms['number'] == storm_number]
    if storm_name is not None:
        storms = storms[storms['name'].str.contains(storm_name.upper())]

    if len(storms) > 0:
        storms['name'] = storms['name'].str.strip()
        storms['class'] = storms['class'].str.strip()
        storms['basin'] = storms['basin'].str.strip()
        storms['source'] = storms['source'].str.strip()
        return storms.iloc[0]
    else:
        message = f'no storms with given info'
        if storm_name is not None:
            message = f'{message} ("{storm_name}")'
        else:
            message = f'{message} ("{basin}{storm_number}")'
        message = f'{message} found in {year}'
        raise ValueError(message)


def atcf_url(
    nhc_code: str = None, file_deck: ATCF_FileDeck = None, mode: ATCF_Mode = None, year=None,
) -> str:
    if nhc_code is not None:
        year = int(nhc_code[4:])

    if mode is None:
        if nhc_code is None:
            raise ValueError('NHC storm code not given')
        entry = get_atcf_entry(basin=nhc_code[:2], storm_number=int(nhc_code[2:4]), year=year)
        if entry['source'] == 'ARCHIVE':
            mode = ATCF_Mode.historical
        else:
            mode = ATCF_Mode.realtime

    if not isinstance(file_deck, ATCF_FileDeck):
        try:
            file_deck = typepigeon.convert_value(file_deck, ATCF_FileDeck)
        except ValueError:
            file_deck = None
    if file_deck is None:
        file_deck = ATCF_FileDeck.ADVISORY

    if not isinstance(mode, ATCF_Mode):
        try:
            mode = typepigeon.convert_value(mode, ATCF_Mode)
        except ValueError:
            mode = None

    if mode == ATCF_Mode.historical:
        if year is None:
            raise ValueError('NHC storm code not given')
        nhc_dir = f'archive/{year}'
        suffix = '.dat.gz'
    else:
        if file_deck == ATCF_FileDeck.ADVISORY:
            nhc_dir = 'aid_public'
            suffix = '.dat.gz'
        elif file_deck == ATCF_FileDeck.BEST:
            nhc_dir = 'btk'
            suffix = '.dat'
        elif file_deck == ATCF_FileDeck.FIXED:
            nhc_dir = 'fix'
            suffix = '.dat'
        else:
            raise NotImplementedError(f'filedeck "{file_deck}" is not implemented')

    url = f'ftp://ftp.nhc.noaa.gov/atcf/{nhc_dir}/'

    if nhc_code is not None:
        url += f'{file_deck.value}{nhc_code.lower()}{suffix}'

    return url


def normalize_atcf_value(value: Any, to_type: type, round_digits: int = None,) -> Any:
    if type(value).__name__ == 'Quantity':
        value = value.magnitude
    if not (value is None or pandas.isna(value) or value == ''):
        if round_digits is not None and issubclass(to_type, (int, float)):
            if isinstance(value, str):
                value = float(value)
            value = round(value, round_digits)
        value = typepigeon.convert_value(value, to_type)
    return value


def read_atcf(
    atcf: Union[PathLike, io.BytesIO, TextIO],
    record_types: List[ATCF_RecordType] = None,
    fort_22: bool = False,
) -> GeoDataFrame:
    """
    read ATCF format

    :param atcf: path or buffered reader
    :param record_types: allowed record types
    :param fort_22: whether to parse `fort.22` fields
    :return: data frame of parsed ATCF data
    """

    if record_types is not None:
        record_types = [
            typepigeon.convert_value(record_type, str) for record_type in record_types
        ]

    if isinstance(atcf, (str, PathLike, Path)):
        atcf = open(atcf)

    lines = (str(line, 'UTF-8') if isinstance(line, bytes) else line for line in atcf)
    lines = (
        (
            entry.strip()
            for entry in line.split(',', maxsplit=len(ATCF_FIELDS) - 1)
            if ~pandas.isna(line)
        )
        for line in lines
    )

    data = DataFrame.from_records(lines, columns=list(ATCF_FIELDS),).astype(
        {field: 'string' for field in ATCF_FIELDS}
    )

    if data['USERDEFINED'].str.contains(',').any():
        if fort_22:
            extra_fields = FORT_22_FIELDS
        else:
            extra_fields = EXTRA_ATCF_FIELDS
        try:
            lines = (str(line, 'UTF-8') if isinstance(line, bytes) else line for line in atcf)
            lines = (
                (
                    entry.strip()
                    for entry in line.split(',', maxsplit=len(extra_fields) - 1)
                    if ~pandas.isna(line)
                )
                for line in lines
            )
            extra_data = DataFrame.from_records(lines, columns=list(extra_fields),).astype(
                {field: 'string' for field in extra_fields}
            )
            data = pandas.concat([data.iloc[:, :-1], extra_data], axis=1)
        except ValueError:
            pass

    if record_types is not None:
        data = data[data['TECH'].isin(record_types)]
        if len(data) == 0:
            raise ValueError(f'no ATCF records found matching "{record_types}"')

    best_track_records = (data['TECH'] == 'BEST') & (
        data.loc[data['TECH'] == 'BEST', 'TECHNUM/MIN'].str.strip().str.len() > 0
    )
    data.loc[best_track_records, 'YYYYMMDDHH'] += data.loc[best_track_records, 'TECHNUM/MIN']
    data.loc[~best_track_records, 'YYYYMMDDHH'] += '00'
    data['YYYYMMDDHH'] = pandas.to_datetime(data['YYYYMMDDHH'], format='%Y%m%d%H%M')

    data.loc[data['LatN/S'].str.endswith('N'), 'LatN/S'] = data['LatN/S'].str.strip('N')
    data.loc[data['LatN/S'].str.endswith('S'), 'LatN/S'] = '-' + data['LatN/S'].str.strip('S')

    data.loc[data['LonE/W'].str.endswith('E'), 'LonE/W'] = data['LonE/W'].str.strip('E')
    data.loc[data['LonE/W'].str.endswith('W'), 'LonE/W'] = '-' + data['LonE/W'].str.strip('W')

    data[['LatN/S', 'LonE/W']] = (
        data[['LatN/S', 'LonE/W']].astype({'LatN/S': float, 'LonE/W': float}, copy=False) / 10
    )

    if pandas.isna(data['RAD']).any():
        raise ValueError(
            'Error: No radial wind information for this storm; '
            'parametric wind model cannot be built.'
        )

    float_fields = [
        field
        for field in (
            'VMAX',
            'MSLP',
            'RAD',
            'RAD1',
            'RAD2',
            'RAD3',
            'RAD4',
            'RADP',
            'RRP',
            'MRD',
            'GUSTS',
            'EYE',
            'MAXSEAS',
            'DIR',
            'SPEED',
            'SEAS',
            'SEAS1',
            'SEAS2',
            'SEAS3',
            'SEAS4',
        )
        if field in data.columns
    ]

    for float_field in float_fields:
        data.loc[
            (data[float_field].str.len() == 0) | pandas.isna(data[float_field]), float_field
        ] = 'NaN'

    data[float_fields] = data[float_fields].astype(float, copy=False)

    data.rename(
        columns={
            **{field: value for field, value in ATCF_FIELDS.items() if field in data.columns},
            **{
                field: value
                for field, value in EXTRA_ATCF_FIELDS.items()
                if field in data.columns
            },
            **{
                field: value
                for field, value in FORT_22_FIELDS.items()
                if field in data.columns
            },
        },
        inplace=True,
    )

    return GeoDataFrame(
        data, geometry=geopandas.points_from_xy(data['longitude'], data['latitude'],)
    )



from datetime import datetime
from enum import Enum
from functools import lru_cache
from os import PathLike
import re
from typing import Any, Dict, List

import geopandas
from geopandas import GeoDataFrame
import pandas
from pandas import DataFrame
import requests
import typepigeon
from typepigeon import convert_value

from stormevents.nhc.storms import nhc_storms


class EventType(Enum):
    """
    https://stn.wim.usgs.gov/STNServices/EventTypes.json
    """

    RIVERINE_FLOOD = 1
    HURRICANE = 2
    DROUGHT = 3
    NOREASTER = 4


class EventStatus(Enum):
    """
    https://stn.wim.usgs.gov/STNServices/EventStatus.json
    """

    ACTIVE = 1
    COMPLETED = 2


class HighWaterMarkType(Enum):
    """
    https://stn.wim.usgs.gov/STNServices/HWMTypes.json
    """

    MUD = 1
    DEBRIS = 2
    VEGETATION_LINE = 3
    SEED_LINE = 4
    STAIN_LINE = 5
    MELTED_SNOW_LINE = 6
    DIRECT_OBSERVATION = 7
    OTHER = 8


class HighWaterMarkQuality(Enum):
    """
    https://stn.wim.usgs.gov/STNServices/HWMQualities.json
    """

    EXCELLENT = 1  # +/- 0.05 ft
    GOOD = 2  # +/- 0.10 ft
    FAIR = 3  # +/- 0.20 ft
    POOR = 4  # +/- 0.40 ft
    VERY_POOR = 5  # +/- > 0.40 ft
    UNKNOWN = 6  # Unknown / Historical


class HighWaterMarkEnvironment(Enum):
    COASTAL = 'Coastal'
    RIVERINE = 'Riverine'


class FloodEvent:
    """
    representation of an arbitrary flood event as defined by the USGS
    """

    URL = 'https://stn.wim.usgs.gov/STNServices/HWMs/FilteredHWMs.json'

    def __init__(self, id: int):
        """
        :param id: USGS event ID

        >>> flood = FloodEvent(182)
        >>> flood.high_water_marks()
                 latitude  longitude            eventName hwmTypeName  ...   hwm_label files siteZone                    geometry
        hwm_id                                                         ...
        22636   32.007730 -81.238270  Irma September 2017   Seed line  ...        HWM1    []      NaN  POINT (-81.23827 32.00773)
        22757   30.510528 -81.460833  Irma September 2017      Debris  ...       HWM 1    []        0  POINT (-81.46083 30.51053)
        22885   30.770560 -81.581390  Irma September 2017   Seed line  ...  GACAM17842    []      NaN  POINT (-81.58139 30.77056)
        22965   31.063150 -81.404540  Irma September 2017      Debris  ...         HWM    []      NaN  POINT (-81.40454 31.06315)
        23052   30.845000 -81.560000  Irma September 2017      Debris  ...  GACAM17840    []      NaN  POINT (-81.56000 30.84500)
        ...           ...        ...                  ...         ...  ...         ...   ...      ...                         ...
        25147   30.018190 -81.859657  Irma September 2017         Mud  ...       HWM01    []      NaN  POINT (-81.85966 30.01819)
        25148   30.097214 -81.891451  Irma September 2017   Seed line  ...      hwm 01    []      NaN  POINT (-81.89145 30.09721)
        25150   30.038222 -81.880928  Irma September 2017   Seed line  ...       HWM01    []      NaN  POINT (-81.88093 30.03822)
        25158   29.720560 -81.506110  Irma September 2017   Seed line  ...         HWM    []      NaN  POINT (-81.50611 29.72056)
        25159   30.097514 -81.794375  Irma September 2017   Seed line  ...       HWM 1    []      NaN  POINT (-81.79438 30.09751)
        [221 rows x 52 columns]
        >>> flood.high_water_marks(quality=['EXCELLENT', 'GOOD'])
                 latitude  longitude            eventName hwmTypeName  ...   hwm_label files siteZone                    geometry
        hwm_id                                                         ...
        22636   32.007730 -81.238270  Irma September 2017   Seed line  ...        HWM1    []      NaN  POINT (-81.23827 32.00773)
        22885   30.770560 -81.581390  Irma September 2017   Seed line  ...  GACAM17842    []      NaN  POINT (-81.58139 30.77056)
        23130   31.034720 -81.640000  Irma September 2017   Seed line  ...        HWM1    []      NaN  POINT (-81.64000 31.03472)
        23216   32.035150 -81.045040  Irma September 2017   Seed line  ...        HWM1    []      NaN  POINT (-81.04504 32.03515)
        23236   32.083650 -81.157520  Irma September 2017   Seed line  ...        HWM1    []      NaN  POINT (-81.15752 32.08365)
        ...           ...        ...                  ...         ...  ...         ...   ...      ...                         ...
        25146   29.992580 -81.851518  Irma September 2017   Seed line  ...      HWM 01    []      NaN  POINT (-81.85152 29.99258)
        25148   30.097214 -81.891451  Irma September 2017   Seed line  ...      hwm 01    []      NaN  POINT (-81.89145 30.09721)
        25150   30.038222 -81.880928  Irma September 2017   Seed line  ...       HWM01    []      NaN  POINT (-81.88093 30.03822)
        25158   29.720560 -81.506110  Irma September 2017   Seed line  ...         HWM    []      NaN  POINT (-81.50611 29.72056)
        25159   30.097514 -81.794375  Irma September 2017   Seed line  ...       HWM 1    []      NaN  POINT (-81.79438 30.09751)
        [138 rows x 52 columns]
        """

        self.id = id

        self.__query = None
        self.__data = None
        self.__error = None

    @classmethod
    def from_name(cls, name: str, year: int = None,) -> 'FloodEvent':
        """
        retrieve high-water mark info from the USGS flood event name

        :param name: USGS flood event name
        :param year: year of flood event
        :return: flood event object
        """

        events = usgs_flood_events(year=year)
        events = events[events['name'] == name]

        if len(events) == 0:
            raise ValueError(f'no event with name "{name}" found')

        if year is not None:
            events = events[events['year'] == year]

        event = events.iloc[0]

        return cls(id=event.name)

    @classmethod
    def from_csv(cls, filename: PathLike) -> 'FloodEvent':
        """
        read a CSV file with high-water mark data

        :param filename: file path to CSV
        :return: flood event object
        """

        data = pandas.read_csv(filename, index_col='hwm_id')
        try:
            instance = cls(id=int(data['event_id'].iloc[0]))
        except KeyError:
            instance = cls.from_name(data['eventName'].iloc[0])
        instance.__data = data
        return instance

    @property
    def id(self) -> int:
        return self.__id

    @id.setter
    def id(self, id: int):
        self.__metadata = usgs_flood_events().loc[id]
        self.__id = self.__metadata.name

    @property
    def name(self) -> str:
        return self.__metadata['name']

    @property
    def year(self) -> int:
        return self.__metadata['year']

    @property
    def description(self) -> str:
        return self.__metadata['description']

    @property
    def event_type(self) -> EventType:
        return typepigeon.convert_value(self.__metadata['event_type'], EventType)

    @property
    def event_status(self) -> EventStatus:
        return typepigeon.convert_value(self.__metadata['event_status'], EventStatus)

    @property
    def coordinator(self) -> str:
        return self.__metadata['coordinator']

    @property
    def instruments(self) -> str:
        return self.__metadata['instruments']

    @property
    def last_updated(self) -> datetime:
        return self.__metadata['last_updated']

    @property
    def last_updated_by(self) -> str:
        return self.__metadata['last_updated_by']

    @property
    def start_date(self) -> datetime:
        return typepigeon.convert_value(self.__metadata['start_date'], datetime)

    @property
    def end_date(self) -> datetime:
        return typepigeon.convert_value(self.__metadata['end_date'], datetime)

    def high_water_marks(
        self,
        us_states: List[str] = None,
        us_counties: List[str] = None,
        hwm_type: HighWaterMarkType = None,
        quality: HighWaterMarkQuality = None,
        environment: HighWaterMarkEnvironment = None,
        survey_completed: bool = None,
        still_water: bool = None,
    ) -> GeoDataFrame:
        """
        :returns: data frame of data for the current parameters

        >>> flood = FloodEvent(23)
        >>> flood.high_water_marks()
                 latitude  longitude eventName                      hwmTypeName  ... approval_id hwm_uncertainty uncertainty                    geometry
        hwm_id                                                                   ...
        14699   38.917360 -75.947890     Irene                           Debris  ...         NaN             NaN         NaN  POINT (-75.94789 38.91736)
        14700   38.917360 -75.947890     Irene                              Mud  ...         NaN             NaN         NaN  POINT (-75.94789 38.91736)
        14701   38.917580 -75.948470     Irene                              Mud  ...         NaN             NaN         NaN  POINT (-75.94847 38.91758)
        14702   38.917360 -75.946060     Irene                       Stain line  ...         NaN             NaN         NaN  POINT (-75.94606 38.91736)
        14703   38.917580 -75.945970     Irene                              Mud  ...         NaN             NaN         NaN  POINT (-75.94597 38.91758)
        ...           ...        ...       ...                              ...  ...         ...             ...         ...                         ...
        41666   44.184900 -72.823970     Irene  Other (Note in Description box)  ...     24707.0             NaN         NaN  POINT (-72.82397 44.18490)
        41667   43.616332 -72.658893     Irene                      Clear water  ...     24706.0             NaN         NaN  POINT (-72.65889 43.61633)
        41668   43.617370 -72.667530     Irene                        Seed line  ...     24705.0             NaN         NaN  POINT (-72.66753 43.61737)
        41670   43.524600 -72.677540     Irene                        Seed line  ...         NaN             NaN         NaN  POINT (-72.67754 43.52460)
        41671   43.534470 -72.672750     Irene                        Seed line  ...         NaN             NaN         NaN  POINT (-72.67275 43.53447)
        [1300 rows x 51 columns]
        """

        if self.__query is None:
            self.__query = HighWaterMarksQuery(
                event_id=self.id,
                event_type=self.event_type,
                event_status=self.event_status,
                us_states=us_states,
                us_counties=us_counties,
                hwm_type=hwm_type,
                quality=quality,
                environment=environment,
                survey_completed=survey_completed,
                still_water=still_water,
            )
        else:
            if us_states is not None:
                self.__query.us_states = us_states
            if us_counties is not None:
                self.__query.us_counties = us_counties
            if hwm_type is not None:
                self.__query.hwm_type = hwm_type
            if quality is not None:
                self.__query.quality = quality
            if environment is not None:
                self.__query.environment = environment
            if survey_completed is not None:
                self.__query.survey_completed = survey_completed
            if still_water is not None:
                self.__query.still_water = still_water

        return self.__query.data

    def __eq__(self, other: 'FloodEvent') -> bool:
        return (
            self.__query is not None
            and other.__query is not None
            and self.__query == other.__query
        )

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(event_id={repr(self.id)})'


class StormFloodEvent(FloodEvent):
    """
    representation of an arbitrary storm flood event as defined by the USGS and NHC
    """

    def __init__(self, name: str, year: int):
        """
        :param name: storm name
        :param year: storm year
        """

        storms = usgs_flood_storms(year=year)
        storm = storms[(storms['nhc_name'] == name.upper().strip()) & (storms['year'] == year)]

        if len(storm) == 0:
            raise ValueError(f'storm "{name} {year}" not found in USGS HWM database')

        super().__init__(id=storm.index[0])


class HighWaterMarksQuery:
    """
    abstraction of an individual query to the USGS Short-Term Network API for high-water marks (HWMs)
    https://stn.wim.usgs.gov/STNServices/Documentation/home
    """

    URL = 'https://stn.wim.usgs.gov/STNServices/HWMs/FilteredHWMs.json'

    def __init__(
        self,
        event_id: int = None,
        event_type: EventType = None,
        event_status: EventStatus = None,
        us_states: List[str] = None,
        us_counties: List[str] = None,
        hwm_type: HighWaterMarkType = None,
        quality: HighWaterMarkQuality = None,
        environment: HighWaterMarkEnvironment = None,
        survey_completed: bool = None,
        still_water: bool = None,
    ):
        """
        :param event_id: USGS event ID
        :param event_type: type of flood event
        :param event_status: whether flood event had completed
        :param us_states: U.S. states in which to query
        :param us_counties: U.S. counties in which to query
        :param hwm_type: HWM type filter
        :param quality: HWM quality filter
        :param environment: HWM environment filter
        :param survey_completed: whether HWM survey should be complete
        :param still_water: HWM still water filter

        >>> query = HighWaterMarksQuery(182)
        >>> query.data
                 latitude  longitude            eventName hwmTypeName  ...   hwm_label files siteZone                    geometry
        hwm_id                                                         ...
        22636   32.007730 -81.238270  Irma September 2017   Seed line  ...        HWM1    []      NaN  POINT (-81.23827 32.00773)
        22757   30.510528 -81.460833  Irma September 2017      Debris  ...       HWM 1    []        0  POINT (-81.46083 30.51053)
        22885   30.770560 -81.581390  Irma September 2017   Seed line  ...  GACAM17842    []      NaN  POINT (-81.58139 30.77056)
        22965   31.063150 -81.404540  Irma September 2017      Debris  ...         HWM    []      NaN  POINT (-81.40454 31.06315)
        23052   30.845000 -81.560000  Irma September 2017      Debris  ...  GACAM17840    []      NaN  POINT (-81.56000 30.84500)
        ...           ...        ...                  ...         ...  ...         ...   ...      ...                         ...
        25147   30.018190 -81.859657  Irma September 2017         Mud  ...       HWM01    []      NaN  POINT (-81.85966 30.01819)
        25148   30.097214 -81.891451  Irma September 2017   Seed line  ...      hwm 01    []      NaN  POINT (-81.89145 30.09721)
        25150   30.038222 -81.880928  Irma September 2017   Seed line  ...       HWM01    []      NaN  POINT (-81.88093 30.03822)
        25158   29.720560 -81.506110  Irma September 2017   Seed line  ...         HWM    []      NaN  POINT (-81.50611 29.72056)
        25159   30.097514 -81.794375  Irma September 2017   Seed line  ...       HWM 1    []      NaN  POINT (-81.79438 30.09751)
        [221 rows x 52 columns]
        >>> query.quality = 'EXCELLENT', 'GOOD'
        >>> query.data
                 latitude  longitude            eventName hwmTypeName  ...   hwm_label files siteZone                    geometry
        hwm_id                                                         ...
        22636   32.007730 -81.238270  Irma September 2017   Seed line  ...        HWM1    []      NaN  POINT (-81.23827 32.00773)
        22885   30.770560 -81.581390  Irma September 2017   Seed line  ...  GACAM17842    []      NaN  POINT (-81.58139 30.77056)
        23130   31.034720 -81.640000  Irma September 2017   Seed line  ...        HWM1    []      NaN  POINT (-81.64000 31.03472)
        23216   32.035150 -81.045040  Irma September 2017   Seed line  ...        HWM1    []      NaN  POINT (-81.04504 32.03515)
        23236   32.083650 -81.157520  Irma September 2017   Seed line  ...        HWM1    []      NaN  POINT (-81.15752 32.08365)
        ...           ...        ...                  ...         ...  ...         ...   ...      ...                         ...
        25146   29.992580 -81.851518  Irma September 2017   Seed line  ...      HWM 01    []      NaN  POINT (-81.85152 29.99258)
        25148   30.097214 -81.891451  Irma September 2017   Seed line  ...      hwm 01    []      NaN  POINT (-81.89145 30.09721)
        25150   30.038222 -81.880928  Irma September 2017   Seed line  ...       HWM01    []      NaN  POINT (-81.88093 30.03822)
        25158   29.720560 -81.506110  Irma September 2017   Seed line  ...         HWM    []      NaN  POINT (-81.50611 29.72056)
        25159   30.097514 -81.794375  Irma September 2017   Seed line  ...       HWM 1    []      NaN  POINT (-81.79438 30.09751)
        [138 rows x 52 columns]
        """

        self.event_id = event_id
        self.event_type = event_type
        self.event_status = event_status
        self.us_states = us_states
        self.us_counties = us_counties
        self.hwm_type = hwm_type
        self.quality = quality
        self.environment = environment
        self.survey_completed = survey_completed
        self.still_water = still_water

        self.__previous_query = self.query
        self.__data = None
        self.__error = None

    @property
    def us_states(self) -> List[str]:
        return self.__us_states

    @us_states.setter
    def us_states(self, us_states: List[str]):
        self.__us_states = typepigeon.convert_value(us_states, [str])

    @property
    def us_counties(self) -> List[str]:
        return self.__us_counties

    @us_counties.setter
    def us_counties(self, us_counties: List[str]):
        self.__us_counties = typepigeon.convert_value(us_counties, [str])

    @property
    def hwm_type(self) -> List[HighWaterMarkType]:
        return self.__hwm_type

    @hwm_type.setter
    def hwm_type(self, hwm_type: HighWaterMarkType):
        if hwm_type is not None:
            self.__hwm_type = convert_value(hwm_type, [HighWaterMarkType])
        else:
            self.__hwm_type = None

    @property
    def quality(self) -> List[HighWaterMarkQuality]:
        return self.__quality

    @quality.setter
    def quality(self, quality: HighWaterMarkQuality):
        if quality is not None:
            self.__quality = convert_value(quality, [HighWaterMarkQuality])
        else:
            self.__quality = None

    @property
    def environment(self) -> List[HighWaterMarkEnvironment]:
        return self.__environment

    @environment.setter
    def environment(self, environment: HighWaterMarkEnvironment):
        if environment is not None:
            self.__environment = convert_value(environment, [HighWaterMarkEnvironment])
        else:
            self.__environment = None

    @property
    def query(self) -> Dict[str, Any]:
        query = {
            'Event': self.event_id,
            'EventType': self.event_type,
            'EventStatus': self.event_status,
            'States': self.us_states,
            'County': self.us_counties,
            'HWMType': self.hwm_type,
            'HWMQuality': self.quality,
            'HWMEnvironment': self.environment,
            'SurveyComplete': self.survey_completed,
            'StillWater': self.still_water,
        }

        for key, value in query.items():
            if key not in ['SurveyComplete', 'StillWater']:
                if isinstance(value, Enum):
                    value = value.value
                elif isinstance(value, List):
                    value = [
                        entry.value if isinstance(entry, Enum) else entry for entry in value
                    ]
                    value = typepigeon.convert_value(value, [str])

                    if len(value) > 0:
                        value = ','.join(value)
                    else:
                        value = None

            query[key] = value

        return query

    @property
    def data(self) -> GeoDataFrame:
        """
        :returns: data frame of data for the current parameters

        >>> query = HighWaterMarksQuery(23)
        >>> query.data
                 latitude  longitude eventName                      hwmTypeName  ... approval_id hwm_uncertainty uncertainty                    geometry
        hwm_id                                                                   ...
        14699   38.917360 -75.947890     Irene                           Debris  ...         NaN             NaN         NaN  POINT (-75.94789 38.91736)
        14700   38.917360 -75.947890     Irene                              Mud  ...         NaN             NaN         NaN  POINT (-75.94789 38.91736)
        14701   38.917580 -75.948470     Irene                              Mud  ...         NaN             NaN         NaN  POINT (-75.94847 38.91758)
        14702   38.917360 -75.946060     Irene                       Stain line  ...         NaN             NaN         NaN  POINT (-75.94606 38.91736)
        14703   38.917580 -75.945970     Irene                              Mud  ...         NaN             NaN         NaN  POINT (-75.94597 38.91758)
        ...           ...        ...       ...                              ...  ...         ...             ...         ...                         ...
        41666   44.184900 -72.823970     Irene  Other (Note in Description box)  ...     24707.0             NaN         NaN  POINT (-72.82397 44.18490)
        41667   43.616332 -72.658893     Irene                      Clear water  ...     24706.0             NaN         NaN  POINT (-72.65889 43.61633)
        41668   43.617370 -72.667530     Irene                        Seed line  ...     24705.0             NaN         NaN  POINT (-72.66753 43.61737)
        41670   43.524600 -72.677540     Irene                        Seed line  ...         NaN             NaN         NaN  POINT (-72.67754 43.52460)
        41671   43.534470 -72.672750     Irene                        Seed line  ...         NaN             NaN         NaN  POINT (-72.67275 43.53447)
        [1300 rows x 51 columns]
        """

        if self.__data is None or self.__previous_query != self.query:
            query = self.query

            if any(
                value is not None
                for key, value in query.items()
                if key not in ['SurveyComplete', 'StillWater']
            ):
                url = self.URL
            else:
                url = 'https://stn.wim.usgs.gov/STNServices/HWMs.json'

            response = requests.get(url, params=query)

            if response.status_code == 200:
                data = DataFrame(response.json())
                self.__error = None
            else:
                self.__error = f'{response.reason} - {response.request.url}'
                raise ValueError(self.__error)

            if len(data) > 0:
                data['survey_date'] = pandas.to_datetime(data['survey_date'], errors='coerce')
                data['flag_date'] = pandas.to_datetime(data['flag_date'], errors='coerce')
                data.loc[data['markerName'].str.len() == 0, 'markerName'] = None
            else:
                data = DataFrame(
                    columns=[
                        'latitude',
                        'longitude',
                        'eventName',
                        'hwmTypeName',
                        'hwmQualityName',
                        'verticalDatumName',
                        'verticalMethodName',
                        'approvalMember',
                        'markerName',
                        'horizontalMethodName',
                        'horizontalDatumName',
                        'flagMemberName',
                        'surveyMemberName',
                        'site_no',
                        'siteDescription',
                        'sitePriorityName',
                        'networkNames',
                        'stateName',
                        'countyName',
                        'siteZone',
                        'sitePermHousing',
                        'site_latitude',
                        'site_longitude',
                        'hwm_id',
                        'waterbody',
                        'site_id',
                        'event_id',
                        'hwm_type_id',
                        'hwm_quality_id',
                        'latitude_dd',
                        'longitude_dd',
                        'survey_date',
                        'elev_ft',
                        'vdatum_id',
                        'vcollect_method_id',
                        'bank',
                        'marker_id',
                        'hcollect_method_id',
                        'hwm_notes',
                        'hwm_environment',
                        'flag_date',
                        'stillwater',
                        'hdatum_id',
                        'hwm_label',
                        'files',
                        'height_above_gnd',
                        'hwm_locationdescription',
                        'flag_member_id',
                        'survey_member_id',
                    ],
                )
            data.set_index('hwm_id', inplace=True)
            self.__data = GeoDataFrame(
                data, geometry=geopandas.points_from_xy(data['longitude'], data['latitude']),
            )
            self.__previous_query = query
        elif self.__error is not None:
            raise ValueError(self.__error)

        return self.__data

    def __eq__(self, other: 'HighWaterMarksQuery') -> bool:
        return self.query == other.query

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}('
            f'event_id={self.event_id}, '
            f'event_type={self.event_type}, '
            f'event_status={self.event_status}, '
            f'us_states={self.us_states}, '
            f'us_counties={self.us_counties}, '
            f'hwm_type={self.hwm_type}, '
            f'quality={self.quality}, '
            f'environment={self.environment}, '
            f'survey_completed={self.survey_completed}, '
            f'still_water={self.still_water}'
            f')'
        )


@lru_cache(maxsize=None)
def usgs_flood_events(
    year: int = None, event_type: EventType = None, event_status: EventStatus = None,
) -> DataFrame:
    """
    this function collects all USGS flood events of the given type and status that have high-water mark data

    https://stn.wim.usgs.gov/STNServices/Events.json

    USGS does not standardize event naming, and they should.
    Year, month, and storm type are not always included.
    The order in which the names are in the response is also not standardized.
    This function applies a workaround to fill in gaps in data.
    USGS should standardize their REST server data.

    :param event_type: type of USGS flood event
    :param year: year of event
    :param event_status: status of USGS flood event
    :return: table of flood events


    >>> usgs_flood_events()
                                                name  year  ...          start_date            end_date
    usgs_id                                                 ...
    7                             FEMA 2013 exercise  2013  ... 2013-05-15 04:00:00 2013-05-23 04:00:00
    8                                          Wilma  2005  ... 2005-10-20 00:00:00 2005-10-31 00:00:00
    9                            Midwest Floods 2011  2011  ... 2011-02-01 06:00:00 2011-08-30 05:00:00
    10                          2013 - June PA Flood  2013  ... 2013-06-23 00:00:00 2013-07-01 00:00:00
    11               Colorado 2013 Front Range Flood  2013  ... 2013-09-12 05:00:00 2013-09-24 05:00:00
    ...                                          ...   ...  ...                 ...                 ...
    311                   2021 August Flash Flood TN  2021  ... 2021-08-21 05:00:00 2021-08-22 05:00:00
    312                    2021 Tropical Cyclone Ida  2021  ... 2021-08-27 05:00:00 2021-09-03 05:00:00
    313                Chesapeake Bay - October 2021  2021  ... 2021-10-28 04:00:00                 NaT
    314      2021 November Flooding Washington State  2021  ... 2021-11-08 06:00:00 2021-11-19 06:00:00
    315          Washington Coastal Winter 2021-2022  2021  ... 2021-11-01 05:00:00 2022-06-30 05:00:00
    [292 rows x 11 columns]
    """

    events = pandas.read_json('https://stn.wim.usgs.gov/STNServices/Events.json')
    events.rename(
        columns={
            'event_id': 'usgs_id',
            'event_name': 'name',
            'event_start_date': 'start_date',
            'event_end_date': 'end_date',
            'event_description': 'description',
            'event_coordinator': 'coordinator',
        },
        inplace=True,
    )
    events.set_index('usgs_id', inplace=True)
    events['start_date'] = pandas.to_datetime(events['start_date'])
    events['end_date'] = pandas.to_datetime(events['end_date'])
    events['last_updated'] = pandas.to_datetime(events['last_updated'])
    events['event_type'] = events['event_type_id'].apply(lambda value: EventType(value).name)
    events['event_status'] = events['event_status_id'].apply(
        lambda value: EventStatus(value).name
    )
    events['year'] = events['start_date'].dt.year
    events = events[
        [
            'name',
            'year',
            'description',
            'event_type',
            'event_status',
            'coordinator',
            'instruments',
            'last_updated',
            'last_updated_by',
            'start_date',
            'end_date',
        ]
    ]

    if event_type is not None:
        event_type = typepigeon.convert_value(event_type, [str])
        events = events[events['event_type'].isin(event_type)]

    if event_status is not None:
        event_status = typepigeon.convert_value(event_status, [str])
        events = events[events['event_status'].isin(event_status)]

    if year is not None:
        year = typepigeon.convert_value(year, [int])
        events = events[events['year'].isin(year)]

    return events


@lru_cache(maxsize=None)
def usgs_flood_storms(year: int = None) -> DataFrame:
    """
    this function collects USGS high-water mark data for storm events and cross-correlates it with NHC storm data

    this is useful if you want to retrieve USGS data for a specific NHC storm code

    :param year: storm year
    :return: table of USGS flood events with NHC storm names

    >>> usgs_flood_storms()
                                                   usgs_name  year  nhc_name  ... last_updated_by          start_date            end_date
    usgs_id                                                                   ...
    8                                                  Wilma  2005     WILMA  ...             NaN 2005-10-20 00:00:00 2005-10-31 00:00:00
    18                                        Isaac Aug 2012  2012     ISAAC  ...            35.0 2012-08-27 05:00:00 2012-09-02 05:00:00
    19                                                  Rita  2005      RITA  ...             NaN 2005-09-23 04:00:00 2005-09-25 04:00:00
    23                                                 Irene  2011     IRENE  ...             NaN 2011-08-26 04:00:00 2011-08-29 04:00:00
    24                                                 Sandy  2012     SANDY  ...             NaN 2012-10-21 04:00:00 2012-10-30 04:00:00
    25                                                Gustav  2008    GUSTAV  ...             NaN 2008-08-31 04:00:00 2008-09-03 04:00:00
    26                                                   Ike  2008       IKE  ...             NaN 2008-09-11 04:00:00 2008-09-12 04:00:00
    119                                              Joaquin  2015   JOAQUIN  ...             NaN 2015-10-01 04:00:00 2015-10-08 04:00:00
    131                                              Hermine  2016   HERMINE  ...             NaN 2016-09-01 04:00:00 2016-09-07 04:00:00
    133                                Isabel September 2003  2003    ISABEL  ...             NaN 2003-09-06 05:00:00 2003-09-30 05:00:00
    135                                 Matthew October 2016  2016   MATTHEW  ...             NaN 2016-10-03 04:00:00 2016-10-30 04:00:00
    180                                      Harvey Aug 2017  2017    HARVEY  ...             NaN 2017-08-24 05:00:00 2017-09-24 05:00:00
    182                                  Irma September 2017  2017      IRMA  ...            35.0 2017-09-03 05:00:00 2017-09-30 05:00:00
    189                                 Maria September 2017  2017     MARIA  ...             1.0 2017-09-17 04:00:00 2017-10-02 04:00:00
    190                                  Jose September 2017  2017      JOSE  ...             1.0 2017-09-18 04:00:00 2017-09-25 04:00:00
    196                                    Nate October 2017  2017      NATE  ...            35.0 2017-10-05 05:00:00 2017-10-14 05:00:00
    281                                     Lane August 2018  2018      LANE  ...            35.0 2018-08-22 05:00:00 2018-09-15 05:00:00
    282                                      Gordon Sep 2018  2018    GORDON  ...            35.0 2018-09-04 05:00:00 2018-10-04 05:00:00
    283                                    Florence Sep 2018  2018  FLORENCE  ...            35.0 2018-09-07 05:00:00 2018-10-07 05:00:00
    284                                       Isaac Sep 2018  2018     ISAAC  ...             3.0 2018-09-11 04:00:00 2018-09-18 04:00:00
    287                                     Michael Oct 2018  2018   MICHAEL  ...             3.0 2018-10-08 04:00:00 2018-10-15 04:00:00
    291                                2019 Hurricane Dorian  2019    DORIAN  ...            36.0 2019-08-28 04:00:00 2019-09-20 04:00:00
    299      1995 South Carolina August Tropical Storm Jerry  1995     JERRY  ...           864.0 1995-08-01 05:00:00 1995-08-31 05:00:00
    300        1999 South Carolina September Hurricane Floyd  1999     FLOYD  ...           864.0 1999-09-01 05:00:00 1999-09-30 05:00:00
    301                                2020 Hurricane Isaias  2020    ISAIAS  ...           864.0 2020-07-31 05:00:00 2020-08-07 05:00:00
    303                     2020 TS Marco - Hurricane  Laura  2020     MARCO  ...           864.0 2020-08-22 05:00:00 2020-08-30 05:00:00
    304                                 2020 Hurricane Sally  2020     SALLY  ...           864.0 2020-09-13 05:00:00 2020-09-20 05:00:00
    305                                 2020 Hurricane Delta  2020     DELTA  ...           864.0 2020-10-06 05:00:00 2020-10-13 05:00:00
    310                          2021 Tropical Cyclone Henri  2021     HENRI  ...           864.0 2021-08-20 05:00:00 2021-09-03 05:00:00
    312                            2021 Tropical Cyclone Ida  2021       IDA  ...           864.0 2021-08-27 05:00:00 2021-09-03 05:00:00
    """

    events = usgs_flood_events(year=year, event_type=EventType.HURRICANE)

    events.rename(columns={'name': 'usgs_name'}, inplace=True)
    events['nhc_name'] = None
    events['nhc_code'] = None

    storms = nhc_storms(tuple(pandas.unique(events['year'])))

    storm_names = pandas.unique(storms['name'].str.strip())
    storm_names.sort()
    for storm_name in storm_names:
        event_storms = events[
            events['usgs_name'].str.contains(storm_name, flags=re.IGNORECASE)
        ]
        for event_id, event in event_storms.iterrows():
            storms_matching = storms[
                storms['name'].str.contains(storm_name, flags=re.IGNORECASE)
                & (storms['year'] == event['year'])
            ]

            for nhc_code, storm in storms_matching.iterrows():
                events.at[event_id, 'nhc_name'] = storm['name']
                events.at[event_id, 'nhc_code'] = storm.name

    return events.loc[
        ~pandas.isna(events['nhc_code']),
        ['usgs_name', 'year', 'nhc_name', 'nhc_code', *events.columns[2:-2],],
    ]
